--- old/usr/src/uts/i86pc/io/pci/pci_common.c	2012-10-01 13:29:58.325252099 -0400
+++ new/usr/src/uts/i86pc/io/pci/pci_common.c	2012-10-01 13:29:57.891016283 -0400
@@ -21,6 +21,8 @@
 
 /*
  * Copyright (c) 2005, 2010, Oracle and/or its affiliates. All rights reserved.
+ *
+ * Copyright (c) 2012, Nexenta Systems, Inc.  All rights reserved.
  */
 
 /*
@@ -200,6 +202,9 @@
 	int			rv;
 	int			behavior;
 	int			cap_ptr;
+	boolean_t		did_pci_config_setup = B_FALSE;
+	boolean_t		did_intr_vec_alloc = B_FALSE;
+	boolean_t		did_msi_cap_set = B_FALSE;
 	uint16_t		msi_cap_base, msix_cap_base, cap_ctrl;
 	char			*prop;
 	ddi_intrspec_t		isp;
@@ -340,6 +345,7 @@
 				    DDI_SUCCESS)
 					return (DDI_FAILURE);
 				i_ddi_set_pci_config_handle(rdip, handle);
+				did_pci_config_setup = B_TRUE;
 			}
 
 			prop = NULL;
@@ -365,6 +371,7 @@
 				}
 			}
 			i_ddi_set_msi_msix_cap_ptr(rdip, cap_ptr);
+			did_msi_cap_set = B_TRUE;
 
 			/*
 			 * Allocate interrupt vectors
@@ -372,8 +379,11 @@
 			(void) (*psm_intr_ops)(rdip, hdlp,
 			    PSM_INTR_OP_ALLOC_VECTORS, result);
 
-			if (*(int *)result == 0)
-				return (DDI_INTR_NOTFOUND);
+			if (*(int *)result == 0) {
+				rv = DDI_INTR_NOTFOUND;
+				goto HANDLE_ALLOC_FAILURE;
+			}
+			did_intr_vec_alloc = B_TRUE;
 
 			/* verify behavior flag and take appropriate action */
 			if ((behavior == DDI_INTR_ALLOC_STRICT) &&
@@ -382,9 +392,8 @@
 				    "pci_common_intr_ops: behavior %x, "
 				    "couldn't get enough intrs\n", behavior));
 				hdlp->ih_scratch1 = *(int *)result;
-				(void) (*psm_intr_ops)(rdip, hdlp,
-				    PSM_INTR_OP_FREE_VECTORS, NULL);
-				return (DDI_EAGAIN);
+				rv = DDI_EAGAIN;
+				goto HANDLE_ALLOC_FAILURE;
 			}
 
 			if (hdlp->ih_type == DDI_INTR_TYPE_MSIX) {
@@ -401,12 +410,8 @@
 						    (void *)rdip,
 						    hdlp->ih_inum));
 
-						(void) (*psm_intr_ops)(rdip,
-						    hdlp,
-						    PSM_INTR_OP_FREE_VECTORS,
-						    NULL);
-
-						return (DDI_FAILURE);
+						rv = DDI_FAILURE;
+						goto HANDLE_ALLOC_FAILURE;
 					}
 				}
 			}
@@ -424,6 +429,19 @@
 		} else
 			return (DDI_FAILURE);
 		break;
+
+HANDLE_ALLOC_FAILURE:
+		if (did_intr_vec_alloc == B_TRUE)
+			(void) (*psm_intr_ops)(rdip, hdlp,
+			    PSM_INTR_OP_FREE_VECTORS, NULL);
+		if (did_msi_cap_set == B_TRUE)
+			i_ddi_set_msi_msix_cap_ptr(rdip, 0);
+		if (did_pci_config_setup == B_TRUE) {
+			(void) pci_config_teardown(&handle);
+			i_ddi_set_pci_config_handle(rdip, NULL);
+		}
+		return (rv);
+
 	case DDI_INTROP_FREE:
 		if (DDI_INTR_IS_MSI_OR_MSIX(hdlp->ih_type) &&
 		    (psm_intr_ops != NULL)) {
